from enum import auto, Enum

class TokenKind(Enum):
    LeftParen = 0
    RightParen = auto()
    Comma = auto()
    OpenTag = auto()
    CloseTag = auto()
    Dot = auto()
    Slash = auto()
    BackSlash = auto()
    # Literals.
    Identifier = auto()
    Number = auto()
    Text = auto()
    # Builtin Variables.
    Black = auto()
    Red = auto()
    Green = auto()
    Yellow = auto()
    Blue = auto()
    Magenta = auto()
    Cyan = auto()
    White = auto()
    Rgb = auto()
    B = auto()
    C = auto()
    I = auto()
    S = auto()
    U = auto()
    X = auto()
    # Keywords.
    Eof = auto()
    Error = auto()

class Token:
    def __init__(
        self,
        kind: TokenKind,
        content: str,
        err_code: int,
        line: int
    ) -> None:
        self.kind = kind
        self.content = content
        self.err_code = err_code
        self.err_code = err_code
        self.line = line


import re
import sys
import os

from typing import TextIO
from .value import C
from .compiler import Compiler

__all__ = ["style", "template", "Compiler"]

class String(TextIO):
    def __init__(self):
        self.s = ""

    def write(self, s: str) -> int:
        self.s = self.s + s
        return len(s)

def _compile(source: str, out: TextIO):
    vars = {
        "green": C.rgb(0, 150, 75),
        "cyan": C.rgb(0, 150, 150)
    }
    compiler = Compiler(source, out, vars)
    compiler.compile()


def style(text: str) -> str:
    vars: dict[str, str] = {}
    out = String()
    compiler = Compiler(text, out, vars)
    compiler.compile()
    return out.s

def template(save: str):
    return lambda text: style(f"{save}{text}")



from typing import Final

class X:
    BLACK: Final[str]   = "\x1b[40m"
    # Red Color.
    RED: Final[str]     = "\x1b[41m"
    # Green Color.
    GREEN: Final[str]   = "\x1b[42m"
    # Yellow Color.
    YELLOW: Final[str]  = "\x1b[43m"
    # Blue Color.
    BLUE: Final[str]    = "\x1b[44m"
    # Magenta Color.
    MAGENTA: Final[str] = "\x1b[45m"
    # Cyan Color.
    CYAN: Final[str]    = "\x1b[46m"
    # White Color.
    WHITE: Final[str]   = "\x1b[47m"

    # RGB Color.
    @staticmethod
    def rgb(r: int, g: int, b: int) -> str:
        return f"\x1b[48;2;{r};{g};{b}m"


from typing import TextIO
from sys import stdout
from scanner import Scanner
from scanner.token import TokenKind

def debug(source: str, out: TextIO):
    scanner = Scanner(source)
    line = -1
    while True:
        token = scanner.scan_token()
        if token.err_code == 0:
            content = token.content
        else:
            content = "Unexpected character."
        if token.line != line:
            out.write("{:4} ".format(token.line))
            line = token.line
        else:
            out.write("   | ")
        out.write("{} '{}'\n".format(token.kind, content))
        if token.kind == TokenKind.Eof:
            break

if __name__ == "__main__":
    debug("""Convenient Terminal Output Styler.

<green><b><u>Usage:</u></b> <cyan><b>ziyy</b> <i>[OPTION] [TEXT]</i></cyan>

<b><u>Options:</u></b></green>
  <cyan><b>-V</b></cyan>, <cyan><b>--version</b></cyan>
          Print version info and exit
  <cyan><b>-f</b></cyan>, <cyan><b>--file</b> \<FILENAME\></cyan>
          Read input from file.
  <cyan><b>-n</b></cyan>, <cyan><b>--no-newline</b></cyan>
          Do not print newline after text.
  <cyan><b>-h</b></cyan>, <cyan><b>--help</b></cyan>
          Print help
""", stdout)


from .state import State
from ..scanner.token import TokenKind as TT
from ..scanner import Scanner
from ..value import *
from typing import TextIO

class ParseException(Exception):
    pass

class Parser:
    def __init__(self, source: str, out: TextIO, variables: dict[str, str]):
        self.scanner = Scanner(source)
        self.out = out
        self.state = State()
        self.variables = variables

    def parse_to_out(self):
        self.out.write(RESET)
        while True:
            token = self.scanner.scan_token()
            match token.kind:
                case TT.Text:
                    self.out.write(token.content)
                case TT.B:
                    self.state.push(token.content, B)
                    self.out.write(B)

                    token = self.scanner.scan_token()
                    if token.kind != TT.CloseTag:
                        raise ParseException
                case TT.I:
                    self.state.push(token.content, I)
                    self.out.write(I)

                    token = self.scanner.scan_token()
                    if token.kind != TT.CloseTag:
                        raise ParseException
                case TT.S:
                    self.state.push(token.content, S)
                    self.out.write(S)

                    token = self.scanner.scan_token()
                    if token.kind != TT.CloseTag:
                        raise ParseException
                case TT.U:
                    self.state.push(token.content, U)
                    self.out.write(U)

                    token = self.scanner.scan_token()
                    if token.kind != TT.CloseTag:
                        raise ParseException
                case TT.C:
                    token = self.scanner.scan_token()
                    match token.kind:
                        case TT.Dot:
                            token = self.scanner.scan_token()
                            match token.kind:
                                case TT.Black:
                                    self.state.push("c", C.BLACK)
                                    self.out.write(C.BLACK)
                                case TT.Blue:
                                    self.state.push("c", C.BLUE)
                                    self.out.write(C.BLUE)
                                case TT.Cyan:
                                    self.state.push("c", C.CYAN)
                                    self.out.write(C.CYAN)
                                case TT.Green:
                                    self.state.push("c", C.GREEN)
                                    self.out.write(C.GREEN)
                                case TT.Magenta:
                                    self.state.push("c", C.MAGENTA)
                                    self.out.write(C.MAGENTA)
                                case TT.Red:
                                    self.state.push("c", C.RED)
                                    self.out.write(C.RED)
                                case TT.White:
                                    self.state.push("c", C.WHITE)
                                    self.out.write(C.WHITE)
                                case TT.Yellow:
                                    self.state.push("c", C.YELLOW)
                                    self.out.write(C.YELLOW)
                                case TT.Rgb:
                                    token = self.scanner.scan_token()
                                    if token.kind != TT.LeftParen:
                                        raise ParseException

                                    token = self.scanner.scan_token()
                                    if token.kind != TT.Number:
                                        raise ParseException
                                    r: int = int(token.content)

                                    token = self.scanner.scan_token()
                                    if token.kind != TT.Comma:
                                        raise ParseException

                                    token = self.scanner.scan_token()
                                    if token.kind != TT.Number:
                                        raise ParseException
                                    g: int = int(token.content)

                                    token = self.scanner.scan_token()
                                    if token.kind != TT.Comma:
                                        raise ParseException

                                    token = self.scanner.scan_token()
                                    if token.kind != TT.Number:
                                        raise ParseException
                                    b: int = int(token.content)

                                    token = self.scanner.scan_token()
                                    if token.kind != TT.RightParen:
                                        raise ParseException

                                    rgb = C.rgb(r, g, b)
                                    self.state.push("c", rgb)
                                    self.out.write(rgb)
                                case _: raise ParseException
                        case _: raise ParseException

                    token = self.scanner.scan_token()
                    if token.kind != TT.CloseTag:
                        raise ParseException
                case TT.X:
                    token = self.scanner.scan_token()
                    match token.kind:
                        case TT.Dot:
                            token = self.scanner.scan_token()
                            match token.kind:
                                case TT.Black:
                                    self.state.push("x", X.BLACK)
                                    self.out.write(X.BLACK)
                                case TT.Blue:
                                    self.state.push("x", X.BLUE)
                                    self.out.write(X.BLUE)
                                case TT.Cyan:
                                    self.state.push("x", X.CYAN)
                                    self.out.write(C.CYAN)
                                case TT.Green:
                                    self.state.push("x", X.GREEN)
                                    self.out.write(X.GREEN)
                                case TT.Magenta:
                                    self.state.push("x", X.MAGENTA)
                                    self.out.write(C.MAGENTA)
                                case TT.Red:
                                    self.state.push("x", X.RED)
                                    self.out.write(C.RED)
                                case TT.White:
                                    self.state.push("x", X.WHITE)
                                    self.out.write(X.WHITE)
                                case TT.Yellow:
                                    self.state.push("x", X.YELLOW)
                                    self.out.write(X.YELLOW)
                                case TT.Rgb:
                                    token = self.scanner.scan_token()
                                    if token.kind != TT.LeftParen:
                                        raise ParseException

                                    token = self.scanner.scan_token()
                                    if token.kind != TT.Number:
                                        raise ParseException
                                    r: int = int(token.content)

                                    token = self.scanner.scan_token()
                                    if token.kind != TT.Comma:
                                        raise ParseException

                                    token = self.scanner.scan_token()
                                    if token.kind != TT.Number:
                                        raise ParseException
                                    g: int = int(token.content)

                                    token = self.scanner.scan_token()
                                    if token.kind != TT.Comma:
                                        raise ParseException

                                    token = self.scanner.scan_token()
                                    if token.kind != TT.Number:
                                        raise ParseException
                                    b: int = int(token.content)

                                    token = self.scanner.scan_token()
                                    if token.kind != TT.RightParen:
                                        raise ParseException

                                    rgb = X.rgb(r, g, b)
                                    self.state.push("x", rgb)
                                    self.out.write(rgb)
                                case _: raise ParseException
                        case _: raise ParseException

                    token = self.scanner.scan_token()
                    if token.kind != TT.CloseTag:
                        raise ParseException
                case TT.Slash:
                    token = self.scanner.scan_token()
                    match token.kind:
                        case TT.B:
                            if self.state.current_tag() != "b":
                                raise ParseException
                            self.state.pop()
                            self.out.write(RESET_B)
                        case TT.I:
                            if self.state.current_tag() != "i":
                                raise ParseException
                            self.state.pop()
                            self.out.write(RESET_I)
                        case TT.S:
                            if self.state.current_tag() != "s":
                                raise ParseException
                            self.state.pop()
                            self.out.write(RESET_S)
                        case TT.U:
                            if self.state.current_tag() != "u":
                                raise ParseException
                            self.state.pop()
                            self.out.write(RESET_U)
                        case TT.C:
                            if self.state.current_tag() != "c":
                                raise ParseException
                            self.state.pop()
                            saved = self.state.current_save()
                            self.out.write(saved)
                        case TT.X:
                            if self.state.current_tag() != "x":
                                raise ParseException
                            self.state.pop()
                            saved = self.state.current_save()
                            self.out.write(saved)
                        case TT.Identifier | TT.Black | TT.Blue | TT.Cyan | TT.Green | TT.Magenta | TT.Red | TT.Rgb | TT.White | TT.Yellow:
                            if self.state.current_tag() != token.content:
                                raise ParseException
                            self.state.pop()
                            saved = self.state.current_save()
                            self.out.write(saved)
                        case _: raise ParseException
                    token = self.scanner.scan_token()
                    if token.kind != TT.CloseTag:
                        raise ParseException
                case TT.Identifier | TT.Black | TT.Blue | TT.Cyan | TT.Green | TT.Magenta | TT.Red | TT.Rgb | TT.White | TT.Yellow:
                    var = self.variables.get(token.content)
                    if var is not None:
                        self.state.push(token.content, var)
                        self.out.write(var)
                    else:
                        raise ParseException

                case TT.Eof:
                    self.out.write(RESET)
                    break
                case _: continue


from .c import C
from .x import X

# Bold Value.
B: str = "\x1b[1m"
# Italics Value.
I: str = "\x1b[3m"
# Underline Value.
U: str = "\x1b[4m"
# Strike through Value.
S: str = "\x1b[9m"

RESET: str   = "\x1b[0m"
RESET_B: str = "\x1b[22m"
RESET_I: str = "\x1b[23m"
RESET_U: str = "\x1b[24m"
RESET_S: str = "\x1b[29m"
RESET_C: str = "\x1b[39m"
RESET_X: str = "\x1b[49m"


from .token import *

def is_alpha(c: str) -> bool:
    return len(c) > 0 and c.isascii() and c.isalpha() or c == '_'

def is_digit(c: str) -> bool:
    return len(c) > 0 and c.isascii() and c.isdigit()

class Scanner:
    def __init__(self, source: str) -> None:
        self.source = source
        self.start = 0
        self.current = 0
        self.line = 1
        self.text_line = 1
        self.text_mode = True
        self.escape = 0

    def is_at_end(self) -> bool:
        return self.current + 1 > len(self.source)

    def advance(self) -> str:
        self.current += 1
        return self.source[self.current - 1]

    def peek(self) -> str:
        try:
            c = self.source[self.current]
        except IndexError:
            return '\0'
        else:
            return c

    def peek_next(self) -> str:
        try:
            c = self.source[self.current + 1]
        except IndexError:
            return '\0'
        else:
            return c

    def make_token(self, kind: TokenKind) -> Token:
        s = self.source[self.start:self.current]
        return Token(kind, s, 0, self.line)

    def error_token(self, code: int) -> Token:
        s = self.source[self.start:self.current]
        return Token(TokenKind.Error, s, code, self.line)

    def text_token(self) -> Token:
        s = self.source[self.start:self.current]
        token = Token(TokenKind.Text, s, 0, self.line)
        self.text_line = self.line
        return token


    def skip_whitespace(self) -> None:
        if self.text_mode:
            return
        while True:
            c = self.peek()
            match c:
                case ' ' | '\r' | '\t':
                    self.advance()
                    continue
                case '\n':
                    self.line += 1
                    self.text_line += 1
                    self.advance()
                    continue
                case _:
                    return

    def check_keyword(
        self,
        start: int,
        length: int,
        rest: str,
        kind: TokenKind,
    ) -> TokenKind:
        s = self.source[(self.start + start):self.current]
        if self.current - self.start == start + length and s == rest:
            return kind
        else:
            return TokenKind.Identifier

    def identifier_kind(self) -> TokenKind:
        if self.current - self.start == 1:
            match self.source[self.start]:
                case 'b': return TokenKind.B
                case 'c': return TokenKind.C
                case 'i': return TokenKind.I
                case 's': return TokenKind.S
                case 'u': return TokenKind.U
                case 'x': return TokenKind.X
                case _: return TokenKind.Identifier
        else:
            match self.source[self.start]:
                case 'b':
                    match self.source[self.start + 1]:
                        case 'l':
                            match self.source[self.start + 2]:
                                case 'a': return self.check_keyword(3, 2, "ck", TokenKind.Black)
                                case 'u': return self.check_keyword(3, 1, "e", TokenKind.Blue)
                                case _: return TokenKind.Identifier
                        case _: return TokenKind.Identifier
                case 'c': return self.check_keyword(1, 3, "yan", TokenKind.Cyan)
                case 'g': return self.check_keyword(1, 4, "reen", TokenKind.Green)
                case 'm': return self.check_keyword(1, 6, "agenta", TokenKind.Magenta)
                case 'r':
                    match self.source[self.start + 1]:
                        case 'e': return self.check_keyword(2, 1, "d", TokenKind.Red)
                        case 'g': return self.check_keyword(2, 1, "b", TokenKind.Rgb)
                        case _: return TokenKind.Identifier
                case 'w': return self.check_keyword(1, 4, "hite", TokenKind.White)
                case 'y': return self.check_keyword(1, 5, "ellow", TokenKind.Yellow)
                case _: return TokenKind.Identifier

    def identifier(self) -> Token:
        while is_alpha(self.peek()) or is_digit(self.peek()):
            self.advance()
        kind = self.identifier_kind()
        return self.make_token(kind)

    def number(self) -> Token:
        while is_digit(self.peek()):
            self.advance()
        return self.make_token(TokenKind.Number)

    def scan_token(self) -> Token:
        if self.escape == 0:
            self.skip_whitespace()
        self.start = self.current
        if self.escape == 2 and self.peek() == '\\':
            self.escape = 1
            self.advance()
            return self.make_token(TokenKind.BackSlash)
        if self.escape == 1:
            self.escape = 0
            self.advance()
            return self.text_token()
        if self.is_at_end():
            return self.make_token(TokenKind.Eof)

        c = self.advance()
        if c == '<':
            self.text_mode = False
            return self.make_token(TokenKind.OpenTag)
        elif c == '>':
            self.text_mode = True
            return self.make_token(TokenKind.CloseTag)

        if self.text_mode:
            while not self.is_at_end():
                if self.peek() == '\n':
                    self.line += 1
                if self.peek() == '\\':
                    self.escape = 2
                    return self.text_token()
                if self.peek() != '<':
                    self.advance()
                else:
                    break
            return self.text_token()

        if is_alpha(c):
            return self.identifier()
        if is_digit(c):
            return self.number()

        match c:
            case '(': return self.make_token(TokenKind.LeftParen)
            case ')': return self.make_token(TokenKind.RightParen)
            case ',': return self.make_token(TokenKind.Comma)
            case '.': return self.make_token(TokenKind.Dot)
            case '/': return self.make_token(TokenKind.Slash)
            case _: return self.error_token(1)


from ..value import *
class State:
    def __init__(self):
        self.tags = [""]
        self.saves = [RESET]

    def push(self, tag: str, string: str):
        l = len(self.saves) - 1
        s = self.saves[l]
        self.saves.append(s + string)
        self.tags.append(tag)

    def pop(self) -> tuple[str, str]:
        a = self.tags.pop()
        b = self.saves.pop()
        return a, b

    def current_tag(self) -> str:
        l = len(self.tags) - 1
        return self.tags[l]

    def current_save(self) -> str:
        l = len(self.saves) - 1
        return self.saves[l]


from .parser import Parser
from typing import TextIO

class Compiler:
    def __init__(self, source: str, out: TextIO, variables: dict[str, str]):
        self.parser = Parser(source, out, variables)

    def compile(self):
        self.parser.parse_to_out()

    def compile_source(self, source: str):
        self.parser.scanner.source = source
        self.compile()


import sys
import os

from .__init__ import _compile

out = sys.stdout

def usage():
    _compile("""Convenient Terminal Output Styler.

<green><b><u>Usage:</u></b> <cyan><b>ziyy</b> <i>[OPTION] [TEXT]</i></cyan>

<b><u>Options:</u></b></green>
  <cyan><b>-V</b></cyan>, <cyan><b>--version</b></cyan>
          Print version info and exit
  <cyan><b>-f</b></cyan>, <cyan><b>--file</b> \<FILENAME\></cyan>
          Read input from file.
  <cyan><b>-n</b></cyan>, <cyan><b>--no-newline</b></cyan>
          Do not print newline after text.
  <cyan><b>-h</b></cyan>, <cyan><b>--help</b></cyan>
          Print help
""", out)

args = sys.argv[1:]
if len(args) < 1:
    usage()
    sys.exit(0)
first = args[0]
if first == "-n" or first == "--no-newline":
    _compile(args[1], out)
elif first == "-f" or first == "--file":
    if len(args) == 1:
        usage()
        sys.exit(1)
    if not os.path.isfile(args[1]):
        usage()
        sys.exit(1)
    file = open(args[1], "r").read()
    _compile(file, out)
elif first == "-V" or first == "--version":
    print("ziyy 2.0.0-beta.1")
elif first == "-h" or first == "--help":
    usage()
    sys.exit(0)
else:
    _compile(first, out)
    out.write("\n")
